import json

import requests
from erp_connector.utils.retryable_session import RetryableSession


class OneIntegrationClient:

    def __init__(self,erp_config):
        self.base_url = f'https://one-integration-{erp_config.env}-http-server.internal.cleartax.co'
        self.auth_token = erp_config.authToken
        self.x_erp_instance_id = erp_config.erpInstanceId

    def get_command(self):
        session = RetryableSession()
        url = f"{self.base_url}/api/one-integration/integration/command/get"
        headers = {
            "accept": "*/*",
            "x-clear-auth-token": self.auth_token,
            "x-erp-instance-id": self.x_erp_instance_id
        }

        try:
            response = session.get(url, headers=headers, timeout=45)
            response.raise_for_status()
            response_data = response.json()
            return response_data
        except requests.exceptions.HTTPError as http_err:
            raise Exception(f"HTTP error occurred: {http_err}")
        except requests.exceptions.ConnectionError as conn_err:
            raise Exception(f"Connection error occurred: {conn_err}")
        except requests.exceptions.Timeout as timeout_err:
            raise Exception(f"Timeout error occurred: {timeout_err}")
        except requests.exceptions.RequestException as req_err:
            raise Exception(f"An error occurred: {req_err}")

    def get_presign(self, file_name, command_id):
        session = RetryableSession()
        url = f"{self.base_url}/api/one-integration/integration/action/pre_sign?file_name={file_name}&command_id={command_id}"
        headers = {
            "accept": "*/*",
            "x-clear-auth-token": self.auth_token,
            "x-erp-instance-id": self.x_erp_instance_id,
            "fileContentType": "application/zip"
        }

        try:
            response = session.get(url, headers=headers, timeout=45)
            response.raise_for_status()
            response_data = response.json()
            return response_data
        except requests.exceptions.HTTPError as http_err:
            raise Exception(f"HTTP error occurred: {http_err}")
        except requests.exceptions.ConnectionError as conn_err:
            raise Exception(f"Connection error occurred: {conn_err}")
        except requests.exceptions.Timeout as timeout_err:
            raise Exception(f"Timeout error occurred: {timeout_err}")
        except requests.exceptions.RequestException as req_err:
            raise Exception(f"An error occurred: {req_err}")

    def upload_to_s3(self, upload_url, file_path):
        session = RetryableSession()
        headers = {
            'Content-Type': 'application/zip'
        }
        payload = open(file_path, 'rb')

        try:
            response = session.put(upload_url, headers=headers, data=payload, timeout=45)
            response.raise_for_status()
            if response.status_code != 200 and response.status_code != 201:
                print("s3 upload failure: ", response.text)
            return None
        except requests.exceptions.HTTPError as http_err:
            raise Exception(f"HTTP error occurred: {http_err}")
        except requests.exceptions.ConnectionError as conn_err:
            raise Exception(f"Connection error occurred: {conn_err}")
        except requests.exceptions.Timeout as timeout_err:
            raise Exception(f"Timeout error occurred: {timeout_err}")
        except requests.exceptions.RequestException as req_err:
            raise Exception(f"An error occurred: {req_err}")

    def post_command(self, presign_url, command_id):
        session = RetryableSession()
        url = f"{self.base_url}/api/one-integration/integration/command/send/result"
        headers = {
            "accept": "*/*",
            'Content-Type': 'application/json',
            "x-clear-auth-token": self.auth_token,
            "x-erp-instance-id": self.x_erp_instance_id
        }
        payload = json.dumps({
            "commandId": command_id,
            "metadata": {
                "s3Details": [
                    {
                        "url": presign_url
                    }
                ]
            }
        })

        try:
            response = session.post(url, headers=headers, data=payload, timeout=45)
            response.raise_for_status()
            response_data = response.json()
            return response_data
        except requests.exceptions.HTTPError as http_err:
            raise Exception(f"HTTP error occurred: {http_err}")
        except requests.exceptions.ConnectionError as conn_err:
            raise Exception(f"Connection error occurred: {conn_err}")
        except requests.exceptions.Timeout as timeout_err:
            raise Exception(f"Timeout error occurred: {timeout_err}")
        except requests.exceptions.RequestException as req_err:
            raise Exception(f"An error occurred: {req_err}")



import json
import os

from erp_connector.clients.one_integration_client import OneIntegrationClient
from erp_connector.db.db_loader import ConnectionLoader
from erp_connector.utils.file_utils import zip_folder
from erp_connector.service.sqlite_service import insert_data, update_data
from erp_connector.config.config_loader import CustomErpConnectorConfig, ConfigLoader


def process_connector(config_path):
    print("running process connector")
    row_id = insert_data()
    status = 'DATA_INITIATED'
    command = None
    command_id = None
    try:
        config_loader = ConfigLoader(config_path)
        custom_erp_connector_config = config_loader.load()
        db_connector = ConnectionLoader.load_connector(custom_erp_connector_config)
        db_connector.connect_db()
        one_integration_client = OneIntegrationClient(custom_erp_connector_config)
        get_command_response = one_integration_client.get_command()
        get_command_result = get_command_response['result']
        command = get_command_result['command']
        command_id = get_command_result['commandId']
        update_data(row_id,command,command_id,status)

        if command == 'DATA_EXTRACTION':
            command_id = get_command_result['commandId']
            json_data = get_command_result['metadata']['erpConfig']['tables']
            query_metadata_list = db_connector.generate_query(None)

            current_path = os.getcwd()
            output_folder_path = os.path.join(current_path, 'output')
            zip_folder_path = os.path.join(current_path, 'output_zip')
            for query_metadata in query_metadata_list:
                table_name = query_metadata["tableName"]
                select_query = query_metadata["query"]
                print(f"select_query: {select_query}")
                json_data = db_connector.fetch_data(select_query)
                local_file_path = f"{output_folder_path}/{table_name}.json"
                os.makedirs(output_folder_path, exist_ok=True)
                os.makedirs(zip_folder_path, exist_ok=True)
                with open(local_file_path, 'w') as json_file:
                    json.dump(json_data, json_file, indent=4)

            zip_file_path = zip_folder(output_folder_path, zip_folder_path, "data.zip")

            presigned_response = one_integration_client.get_presign("data.zip", command_id)
            presigned_url = presigned_response['result']['payload']
            one_integration_client.upload_to_s3(presigned_url, zip_file_path)
            post_response = one_integration_client.post_command(presigned_url, command_id)
            print(f"{post_response['message']}")
        status = 'PROCESSED'
    except Exception as e:
        print(f"An error occurred: {e}")
        status = 'ERROR'
    finally:
        update_data(row_id,command,command_id,status)
        db_connector.close_connection()




import sqlite3
from datetime import datetime, timedelta


def create_table():
    conn = sqlite3.connect('scheduler_data.db')
    c = conn.cursor()
    c.execute('''CREATE TABLE IF NOT EXISTS scheduler_data
                 (id INTEGER PRIMARY KEY,
                 started_at TEXT,
                 updated_at TEXT,
                 command TEXT,
                 command_id TEXT,
                 status TEXT)''')
    conn.commit()
    conn.close()


def insert_data():
    start_time = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    conn = sqlite3.connect('scheduler_data.db')
    c = conn.cursor()
    c.execute("INSERT INTO scheduler_data (started_at,updated_at) VALUES (?, ?)",
              (start_time, start_time))
    conn.commit()
    row_id = c.lastrowid
    conn.close()
    return row_id


def update_data(row_id, command,command_id, status):
    conn = sqlite3.connect('scheduler_data.db')
    updated_at = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    c = conn.cursor()
    c.execute("UPDATE scheduler_data SET updated_at = ?, command = ?, command_id = ?, status = ? WHERE id = ?",
              (updated_at, command,command_id, status, row_id))
    conn.commit()
    conn.close()


def delete_old_records():
    conn = sqlite3.connect('scheduler_data.db')
    c = conn.cursor()
    one_month_ago = datetime.now() - timedelta(days=30)
    c.execute("DELETE FROM scheduler_data WHERE start_time < ?", (one_month_ago,))
    conn.commit()
    conn.close()



from sqlglot import parse


def _generate_join_clause(joins):
    clauses = []
    for join in joins:
        join_table = join["name"]
        join_alias = join["alias"]
        join_conditions = " AND ".join([
            f"{cond['left']['alias']}.{cond['left']['name']} = {cond['right']['alias']}.{cond['right']['name']}"
            for cond in join["on"]
        ])
        clause = f"JOIN {join_table} AS {join_alias} ON {join_conditions}"
        clauses.append(clause)
    return " ".join(clauses)


def _generate_expr(expr):
    if "column" in expr:
        column = f"{expr['column']['alias']}.{expr['column']['name']}"
        operator = expr["operator"]
        value = expr["value"][0]  # Assuming only one value in the list
        return f"{column} {operator} '{value}'"
    elif "booleanOperator" in expr:
        return expr["booleanOperator"]


def _generate_order_by_clause(order_by):
    clauses = [f"{item['column']['alias']}.{item['column']['name']} {item.get('order', 'ASC')}" for item in order_by]
    return "ORDER BY " + ", ".join(clauses) if clauses else ""


def _generate_where_clause(conditions):
    clauses = []
    for cond_metadata in conditions:
        if "listOfExpr" in cond_metadata:
            sublist = [_generate_expr(expr) for expr in cond_metadata["listOfExpr"]]
            clauses.append("(" + " ".join(sublist) + ")")
        elif "expr" in cond_metadata and "column" in cond_metadata["expr"]:
            expr = cond_metadata["expr"]
            clauses.append(_generate_expr(expr))
    return "WHERE " + " AND ".join(clauses) if clauses else ""


def generate_data_query(json_data):
    queries = []
    for entity in json_data:
        select_clause = ", ".join([f"{item['name']}" for item in entity["select"]])
        from_table = entity["fromTable"]["name"]
        join_clause = _generate_join_clause(entity.get("join", []))
        where_clause = _generate_where_clause(entity.get("where", []))
        order_by_clause = _generate_order_by_clause(entity.get("order_by", []))

        # Construct the SQL query string
        sql_query = f"SELECT {select_clause} FROM {from_table} {join_clause} {where_clause} {order_by_clause}"

        # Parse the SQL query string to validate and normalize it
        parsed_query = parse(sql_query)
        queries.append({"tableName": from_table, "query": str(parsed_query)})
    return queries



import json

from erp_connector.utils.singleton import SingletonMeta


class CustomErpConnectorConfig(metaclass=SingletonMeta):
    def __init__(self, config_dict):
        self._config = config_dict

    @property
    def dbType(self):
        return self._config.get("dbType")

    @property
    def connectionDetails(self):
        return self._config.get("connectionDetails")

    @property
    def env(self):
        return self._config.get("env")

    @property
    def authToken(self):
        return self._config.get("authToken")

    @property
    def erpInstanceId(self):
        return self._config.get("erpInstanceId")


class ConfigLoader(metaclass=SingletonMeta):

    def __init__(self, config_path):
        self.config_path = config_path

    def load(self):
        with open(self.config_path, 'r') as file:
            config_dict = json.load(file)
        return CustomErpConnectorConfig(config_dict)



import json
from datetime import date
from decimal import Decimal

import mysql.connector
from erp_connector.db.db_connector import DBConnector
from erp_connector.utils.mysql_query_utils import generate_data_query


class MySqlConnector(DBConnector):

    def __init__(self, connection_details):
        super().__init__(connection_details)
        self.connection = None

    def connect_db(self):
        try:
            self.connection = mysql.connector.connect(
                host=self.connection_details['host'],
                user=self.connection_details['user'],
                password=self.connection_details['password'],
                database=self.connection_details['database']
            )
            print(f"Connected to MySQL database: {self.connection_details['database']}")
        except mysql.connector.Error as e:
            print(f"Error connecting to MySQL database: {e}")

    def fetch_data(self, query):
        try:
            cursor = self.connection.cursor(dictionary=True)  # Use dictionary cursor to fetch rows as dictionaries
            cursor.execute(query)
            data = cursor.fetchall()

            # Convert Decimal objects to float
            for row in data:
                for key, value in row.items():
                    if isinstance(value, Decimal):
                        row[key] = float(value)
                    elif isinstance(value, date):
                        row[key] = value.strftime('%Y-%m-%d')

            cursor.close()
            return json.dumps(data)  # Convert data to JSON string
        except mysql.connector.Error as e:
            print(f"Error fetching data from MySQL database: {e}")
            return None  # Return None if there's an error

    def insert_data(self, update_query):
        try:
            cursor = self.connection.cursor()
            cursor.execute(update_query)
            self.connection.commit()
            cursor.close()
            print("Data inserted into MySQL database successfully")
        except mysql.connector.Error as e:
            self.connection.rollback()
            print(f"Error inserting data into MySQL database: {e}")

    def generate_query(self, json_data):
        return generate_data_query(json_data)

    def close_connection(self):
        if self.connection:
            self.connection.close()
            print("Connection to MySQL database closed")



import argparse
from apscheduler.schedulers.background import BackgroundScheduler
from erp_connector.service.connector_service import process_connector
import time
from erp_connector.service.sqlite_service import create_table


def run_scheduler(config_path):
    create_table()
    scheduler = BackgroundScheduler(daemon=True)
    scheduler.add_job(process_connector, 'interval', minutes=1, max_instances=1,args=(config_path,))
    scheduler.start()

    try:
        while True:
            time.sleep(1)
    except (KeyboardInterrupt, SystemExit):
        scheduler.shutdown()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='ERP Connector Scheduler')
    parser.add_argument('config_path', type=str, help='Path to configuration file')
    args = parser.parse_args()
    run_scheduler(args.config_path)



import os
import zipfile


def zip_folder(folder_path, zip_path, zip_name):
    zip_path = os.path.join(zip_path, zip_name)

    with zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED) as zipf:
        for file_name in os.listdir(folder_path):
            file_path = os.path.join(folder_path, file_name)
            if os.path.isfile(file_path):
                zipf.write(file_path, arcname=file_name)
    return zip_path



import json

from erp_connector.db.mssql_connector import MsSqlConnector
from erp_connector.db.mysql_connector import MySqlConnector
from erp_connector.db.postgresql_connector import PostgresSQLConnector


class ConfigLoader:
    def __init__(self, path):
        self.path = path

    def load(self):
        with open(self.path, 'r') as file:
            config = json.load(file)
        return config


class ConnectionLoader:
    @staticmethod
    def load_connector(custom_erp_connector_config):
        connector = ConnectorFactory.create_connector(custom_erp_connector_config)
        return connector


class ConnectorFactory:
    @staticmethod
    def create_connector(erp_connector_config):
        db_type = erp_connector_config.dbType
        connection_details = erp_connector_config.connectionDetails
        if db_type == 'mysql':
            print("creating mysql connector")
            return MySqlConnector(connection_details)
        elif db_type == 'postgresql':
            print("creating postgresql connector")
            return PostgresSQLConnector(connection_details)
        elif db_type == 'mssql':
            print("creating mssql connector")
            return MsSqlConnector(connection_details)
        else:
            raise ValueError("Invalid db connector config: dbName must be 'mysql' or 'postgresql'")



from erp_connector.utils.singleton import SingletonMeta


class InProcessBatchManager(metaclass=SingletonMeta):
    def __init__(self):
        print("init InProcessBatch")
        self._batch_id = None
        self._elt_client = None
        self._is_simulation = None
        self._metadata = {}

    @property
    def batch_id(self):
        return self._batch_id

    @batch_id.setter
    def batch_id(self, value):
        self._batch_id = value

    @property
    def elt_client(self):
        return self._elt_client

    @elt_client.setter
    def elt_client(self, value: str):
        self._elt_client = value

    @property
    def is_simulation(self):
        return self._is_simulation

    @is_simulation.setter
    def is_simulation(self, value):
        self._is_simulation = value

    @property
    def metadata(self):
        return self._metadata

    def set_metadata(self, key, value):
        self._metadata[key] = value

    def clean_up(self):
        self._batch_id = None
        self._elt_client = None
        self._is_simulation = None



from abc import ABC, abstractmethod


class DBConnector(ABC):
    def __init__(self, connection_details):
        self.connection_details = connection_details
        self.connection = None

    @abstractmethod
    def connect_db(self):
        pass

    @abstractmethod
    def fetch_data(self, query):
        pass

    @abstractmethod
    def insert_data(self, insert_query):
        pass

    @abstractmethod
    def generate_query(self, json_data):
        pass

    @abstractmethod
    def close_connection(self):
        pass



import psycopg2
from erp_connector.db.db_connector import DBConnector


class PostgresSQLConnector(DBConnector):

    def __init__(self, connection_details):
        super().__init__(connection_details)
        self.connection = None

    def connect_db(self):
        try:
            self.connection = psycopg2.connect(
                host=self.connection_details['host'],
                user=self.connection_details['user'],
                password=self.connection_details['password'],
                database=self.connection_details['database']
            )
            print(f"Connected to PostgreSQL database: {self.connection_details['database']}")
        except psycopg2.Error as e:
            print(f"Error connecting to PostgreSQL database: {e}")

    def fetch_data(self, query):
        try:
            cursor = self.connection.cursor()
            cursor.execute(query)
            data = cursor.fetchall()
            cursor.close()
            return data
        except psycopg2.Error as e:
            print(f"Error fetching data from PostgreSQL database: {e}")

    def insert_data(self, update_query):
        try:
            cursor = self.connection.cursor()
            cursor.execute(update_query)
            self.connection.commit()
            cursor.close()
            print("Data inserted into PostgreSQL database successfully")
        except psycopg2.Error as e:
            self.connection.rollback()
            print(f"Error inserting data into PostgreSQL database: {e}")

    def close_connection(self):
        if self.connection:
            self.connection.close()
            print("Connection to PostgreSQL database closed")



import os
import time

from erp_connector.utils.custlogging import LoggerProvider
from requests import Session
from requests.exceptions import HTTPError, ConnectionError, ProxyError, Timeout, ConnectTimeout, SSLError

from ..exceptions import HttpTransientException

logger = LoggerProvider().get_logger(os.path.basename(__file__))


class RetryableSession(Session):
    """
    Retries for 500, 501, 502, 503, 504 and network failures
    Can add/remove custom error codes as per requirements
    """

    def request(self, method, url, max_retry=5, backoff_factor=15, **kwargs):
        counter = 0
        while counter < max_retry:
            counter += 1
            try:
                r = super(RetryableSession, self).request(method, url, **kwargs)
                if r.status_code == 200:
                    return r
                elif r.status_code in [500, 501, 502, 503, 504]:
                    logger.warn(f"Received error response from api: {r.status_code}")
                    raise HTTPError()
                else:
                    raise Exception(
                        f"Unrecoverable status code {r.status_code} received from {method} {url}. Failing the batch...")
            except (HTTPError, SSLError, ConnectionError, ConnectTimeout, ProxyError, Timeout) as e:
                if counter < max_retry:
                    # backoff for 15s, 30s, 45s and exit
                    delay = backoff_factor * counter
                    logger.warn(f"Received retryable error from {method} {url}, retry count: {counter}")
                    time.sleep(delay)
                else:
                    logger.error(f"Failed to get data from {method} {url} {kwargs}")
                    raise HttpTransientException



import sys
import traceback


class WorkerRetryableException(Exception):
    """Exception raised for errors that are potentially recoverable and the action can be retried."""

    def __init__(self, original_exception: Exception, message="An error occurred that can be retried."):
        self.original_exception = original_exception
        full_message = f"WorkerRetryableException: {message} Original error: {str(original_exception)}"
        super().__init__(full_message)


class WorkerNonRetryableException(Exception):
    """Exception raised for errors that are not recoverable and the action should not be retried."""

    def __init__(self, original_exception: Exception, message="An error occurred that cannot be retried."):
        self.original_exception = original_exception
        full_message = f"WorkerNonRetryableException: {message} Original error: {str(original_exception)}"
        super().__init__(full_message)


def get_formatted_exception(e: Exception) -> str:
    exc_type, exc_value, exc_traceback = sys.exc_info()
    formatted_traceback = ''.join(traceback.format_tb(exc_traceback))
    lineno = exc_traceback.tb_lineno
    exception_name = e.__class__.__name__
    return f"""{exception_name} processing message  
                                            occurred line {lineno}: {e}
                                            Traceback: {formatted_traceback}"""


class HttpTransientException(Exception):
    """Raised when http server timeouts"""
    pass



import logging
import sys

from erp_connector.utils.InProcessBatch import InProcessBatchManager
from erp_connector.utils.singleton import SingletonMeta


class OneLineExceptionHandler(logging.Formatter):
    def format(self, record):
        if record.exc_info:
            record.msg = repr(super().formatException(record.exc_info))
            record.exc_info = None
            record.exc_text = None
        result = super().format(record)
        return result


FORMATTER = OneLineExceptionHandler(
    "%(levelname)s [%(filename)s:%(lineno)s]:  %(message)s Context: [ %(elt_client)s : %(batch_id)s] ")


class LoggerProvider(metaclass=SingletonMeta):

    def __init__(self):
        print("init logger")
        self.loggers = {}
        self.global_log_level = self.get_global_log_level("INFO")

    def get_logger(self, logger_name: str, log_level=None, add_filter: bool = True):
        log_level = log_level if log_level else self.global_log_level

        if self.loggers.get(logger_name):
            return self.loggers.get(logger_name)
        else:
            logger = logging.getLogger(logger_name)
            logger.setLevel(log_level)
            logger.addHandler(self.get_console_handler())
            if add_filter:
                logger.addFilter(LogFilter())
            else:
                logger.addFilter(LogFilterDummy())
            logger.propagate = False
            self.loggers[logger_name] = logger
            return logger

    @staticmethod
    def get_console_handler():
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setFormatter(FORMATTER)
        return console_handler

    @staticmethod
    def get_global_log_level(level):
        level = level.upper() if level else ''
        log_levels = {
            'ERROR': logging.ERROR,
            'ERR': logging.ERROR,
            'WARNING': logging.WARNING,
            'WARN': logging.WARNING,
            'DEBUG': logging.DEBUG
        }
        return log_levels.get(level, logging.INFO)


class LogFilterDummy(logging.Filter):

    def __init__(self):
        super().__init__()

    def filter(self, record) -> bool:
        record.elt_client = None
        record.batch_id = None
        return True


class LogFilter(logging.Filter):

    def __init__(self):
        super().__init__()
        self.in_process_batch = InProcessBatchManager()

    def filter(self, record) -> bool:
        record.elt_client = self.in_process_batch.elt_client
        record.batch_id = self.in_process_batch.batch_id

        return True



#!/usr/bin/env python
from setuptools import setup, find_packages
import os

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md')) as f:
    long_description = f.read()

package_name = "custom_erp_connector"

package_version = "0.3"
description = """custom-erp-connector"""

setup(
    name=package_name,
    version=package_version,
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Dinesh Sharma",
    author_email="dinesh.sharma@clear.in",
    url="https://github.com/ClearTax/custom-erp-connector",
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'run_scheduler = scheduler:run_scheduler'
        ]
    },
    install_requires=[
        "mysql-connector-python==8.3.0",
        "psycopg2==2.9.9",
        "requests==2.31.0",
        "APScheduler==3.10.4",
        "pyodbc==5.1.0",
        "sqlparse==0.4.3",
        "pandas==1.5.1",
        "SQLAlchemy==1.4.46",
        "python-dateutil==2.8.2",
        "h2==4.1.0",
        "simplejson==3.18.3",
        "sqlglot==18.7.0"
    ]
)
